from klefki.types.algebra.utils import randfield
from klefki.zkp.sigma import NIZK
from klefki.types.algebra.concrete import FiniteFieldCyclicSecp256k1 as CF
from klefki.types.algebra.concrete import EllipticCurveGroupSecp256k1 as Curve
from hashlib import sha256

class NIZKSchnoor(NIZK):
    """
    ref: https://crypto.stanford.edu/cs355/19sp/lec5.pdf
         https://tools.ietf.org/html/rfc8235
    """

    @staticmethod
    def proof(x: CF):
        x = CF(x)
        r = randfield(CF)
        G = Curve.G
        u = G @ r
        h = G @ x
        c = CF(int(sha256(
            str(G.x.value).encode() +
            str(G.y.value).encode() +
            str(h.x.value).encode() +
            str(h.y.value).encode() +
            str(u.x.value).encode() +
            str(u.y.value).encode()
        ).hexdigest(), 16))

        z = r + c * x

        return (u, h, c, z)

    @staticmethod
    def verify(u, h, c, z):
        G = Curve.G
        return (c == CF(int(sha256(
            str(G.x.value).encode() +
            str(G.y.value).encode() +
            str(h.x.value).encode() +
            str(h.y.value).encode() +
            str(u.x.value).encode() +
            str(u.y.value).encode()
        ).hexdigest(), 16))) and (
            G@z == u + h@c
        )



from klefki.numbers import invmod
from klefki.numbers import rsa_lambda, rsa_phi
from klefki.numbers import lcm, length
from klefki.numbers import crt
from klefki.crypto.paillier import Paillier
from klefki.types.algebra.meta import field
from klefki.numbers.primes import generate_prime
from klefki.types.algebra.utils import randfield
from functools import lru_cache
from math import factorial, gcd
import damgard_jurik as ts_dj


def damgard_jurik_reduce(a: int, n: int, s=1) -> int:
    """Computes i given a = (1 + n)^i (mod n^(s+1)).
    :param a: The integer a in the above equation.
    :param s: The integer s in the above equation.
    :param n: The integer n in the above equation.
    :return: The integer i in the above equation.
    """
    def L(b: int) -> int:
        assert (b - 1) % n == 0
        return (b - 1) // n

    @lru_cache(int(s))
    def n_pow(p: int) -> int:
        return n ** p

    @lru_cache(int(s))
    def fact(k: int) -> int:
        return factorial(k)

    i = 0
    for j in range(1, s + 1):

        t_1 = L(a % n_pow(j + 1))
        t_2 = i

        for k in range(2, j + 1):
            k = k

            i = i - 1
            t_2 = t_2 * i % n_pow(j)
            t_1 = t_1 - (t_2 * n_pow(k - 1) * invmod(fact(k), n_pow(j))) % n_pow(j)

        i = t_1
    return i


class DJPaillier(Paillier):

    def __init__(self, P, Q, s=1, strict=False):
        '''
        if strict == True:
           use Carmichael Number
        else:
           use Euler Number
        '''
        N = P * Q
#        Lam = lcm(P-1, Q-1)
        G = field(N**s, "G") # n ** s == n if s = 1
        # multiplicative group
        MG = field(N ** (s+1), "N^{s+1}") # n ** (s +1 ) == n2 in pailer case
        H = field(N, "H")
        # https://crypto.stackexchange.com/questions/29591/lcm-versus-phi-in-rsa
        if strict:
            LG = field(rsa_lambda(P, Q), "PhiGroup")
        else:
            LG = field(rsa_phi(P, Q), "PhiGroup")

        j = generate_prime(length(P))
        assert gcd(j, N) == 1
        x = randfield(H)
        g = MG((MG(1 + N) ** j) * x)

        d = crt(a_list=[0, 1], n_list=[LG.P, G.P])
        assert d % G.P == 1
        assert d % LG.P == 0

        self.s = s
        self.privkey = d
        self.N = N
        self.G = g
        self.pubkey = (N, g)

    @staticmethod
    def cal_privkey(P, Q, s=1):
        return crt(a_list=[0, 1], n_list=[(P-1) * (Q-1), N ** s])

    @classmethod
    def encrypt(cls, m, pub, s=1, r=None):
        N, G = pub

        if hasattr(m, "value"):
            m = m.value
        if not r:
            r = randfield(G.functor)
        else:
            r = G.functor(r)
        return G**m * r**(N**s)


    @classmethod
    def decrypt(cls, c, priv, pub, s=1):
        N, G = pub
        d = priv
        F = field(N**s, "N^s")
        return F(damgard_jurik_reduce((c ** d).value, N, s)) * ~F(damgard_jurik_reduce((G ** d).value, N, s))

    def E(self, m, pub=None, s=None, r=None):
        if not s:
            s = self.s or 1
        return self.encrypt(m, pub or self.pubkey, s=s, r=r)


    def D(self, c, priv=None, pub=None, s=None):
        if not s:
            s = self.s or 1
        return self.decrypt(c, priv or self.privkey, pub or self.pubkey, s=s)



from typing import Tuple, TypeVar, Iterable
from klefki.numbers import invmod
from functools import lru_cache

__all__ = [
    'extended_euclidean_algorithm',
    'double_and_add_algorithm'
]

T = TypeVar('T')

def bits(n: int) -> Iterable:
    """
    Generates the binary digits of n, starting
    from the least significant bit.
    bits(151) -> 1, 1, 1, 0, 1, 0, 0, 1
    """
    def _trans(n):
        while n:
            yield n & 1
            n >>= 1

    return list(_trans(n))


def bits_little_endian_from_bytes(s):
    return "".join([bin(ord(x))[2:].rjust(8,'0')[::-1] for x in s.decode()])

def bytes_from_bits_little_endian(s):
    return bytes([int(s[i:i+8][::-1], 2) for i in range(0, len(s), 8)])

def bytes_mul(a, b, s=32):
    t = [0] * 2* s
    a = a.to_bytes(s, "big")[::-1]
    b = b.to_bytes(s, "big")[::-1]

    for i in range(0, s):
        C = 0
        for j in range(0, s):
            (C, S) = (t[i+j] + a[j] * b[i] + C).to_bytes(2, "big")
            t[i+j] = S
        t[i+s] = C
    return t[::-1]


@lru_cache(maxsize=None)
def montgomery_property(r, n):
    r_inv = invmod(r, n)
    return (r * r_inv -1) // n

def mon_pro(a_, b_, r, n, n_):
    t = a_ * b_
    m = (t * n_) % r
    u_ = (t + m * n) // r
    if u_ >= n:
        return u_ - n
    else:
        return u_

def montgomery_mul(a, b, n):
    r = 2 ** (n.bit_length() +1)
    n_ = montgomery_property(r, n)
    a_ = (a * r) % n
    b_ = (b * r) % n
    u_ = mon_pro(a_, b_, r, n, n_)
    return mon_pro(u_, 1, r, n, n_)



def CIOS(a, b, P, N=32):
    """
    # a[i], b[i], q[i] is the ith word of the numbers a,b,q
    # N is the number of machine words needed to store the modulus q
    Ref: https://hackmd.io/@zkteam/modular_multiplication
         https://www.microsoft.com/en-us/research/wp-content/uploads/1998/06/97Acar.pdf
    """

    a = a.to_bytes(N, "little")
    b = b.to_bytes(N, "little")
    q = P.to_bytes(N, "little")
    # R is set to the smallest power of two exceeding q that falls on a computer word boundary.
    # For example, if q is 381 bits then R=2^{6×64}=2^384 on a 64-bit architecture.
    # For klefki, if q is 256 bits then R=2^{32x8}=2^256 on klefki 8-bit(byte) implementation
    R = 2 ** (N*8)
    # D is the word size. For example, on a 64-bit architecture D is 2^64
    # For bytes, D is 2^8 = 256
    D = 2 ** 8
    # q'[0] is the lowest word of the number −q^{−1} mod R.
    #q_0 = list(int((-P**-1) % R).to_bytes(64, "little"))[0]
    FieldR = type("Field", (FiniteField, ), dict(P=R))
    q_0 = ((-(~FieldR(F.P))).value).to_bytes(N, "little")[0]
    # t is a temporary array of size N+2
    t = [0] * (N + 2)

    for i in range(0, N):
        # cal multi
        C = 0
        for j in range(0, N):
            (C, t[j]) = (t[j] + a[j] * b[i] + C).to_bytes(2, "big")
        (t[N+1], t[N]) = (t[N] + C).to_bytes(2, "big")

        # cal mod
        C = 0
        m = (t[0] * q_0) % D
        (C, _) = (t[0] + m * q[0]).to_bytes(2, "big")
        for j in range(1, N):
            (C, t[j-1]) = (t[j] + m * q[j] + C).to_bytes(2, "big")
        (C, t[N-1]) = (t[N] + C).to_bytes(2, "big")
        t[N] = t[N+1] + C
    return int.from_bytes(t, "little")



def complex_truediv_algorithm(x: complex, y: complex, f: T) -> T:
    a = f(x.real)
    b = f(x.imag)
    c = f(y.real)
    d = f(y.imag)

    return f(complex(
        (a*c + b*d) / (c**2 + d**2),
        (b*c - a*d) / (c**2 + d**2)
    ))


def extended_euclidean_algorithm(a: int, b: int) -> Tuple[int, int, int]:
    '''
    Returns a three-tuple (gcd, x, y) such that
    a * x + b * y == gcd, where gcd.

    This function implements the extended Euclidean
    algotithm and runs in O(log b) in the worst case
    '''
    s, t, r = 0, 1, b
    old_s, old_t, old_r = 1, 0, a

    while r != 0:
        quoient = old_r // r
        old_r, r = r, old_r - quoient * r
        old_s, s = s, old_s - quoient * s
        old_t, t = t, old_t - quoient * t
    return old_r, old_s, old_t


def double_and_add_algorithm(times: int, x: T, init: T) -> T:
    """
    Returns the result of n * x, computed using
    the double and add algorithm.
    """

    result = init
    addend = x

    for bit in bits(times):
        if bit == 1:
            result = addend + result
        addend = addend + addend

    return result


def newton_iterator_sqrt(x: T):
    if (x.value == 0): return x
    last = x.__class__(0)
    res = x.__class__(1)
    while res != last:
        last = res
        res = (res + x / res) / x.__class__(2)
    return res



from klefki.types.algebra.concrete import EllipticCurveCyclicSubgroupSecp256k1 as Curve
from klefki.types.algebra.concrete import FiniteFieldCyclicSecp256k1 as F
from klefki.types.algebra.utils import randfield
from klefki.types.algebra.meta import field
from klefki.types.algebra.isomorphism import bijection
from klefki.utils import int_to_byte, byte_to_int


def map_to_curve(m, G=Curve.G, bias=1):
    try:
        sm = str(getattr(m, "value", m))
        im = int(str(bias) + sm)
        if im > G.x.functor.P:
            raise Exception("secret too large")
        return Curve.lift_x(G.x.functor(im))
    except AssertionError:
        return map_to_curve(m, G=G, bias=bias+1)

@bijection(map_to_curve)
def map_from_curve(M):
    return int(str(M.x.value)[1:])



class ElGamal():

    def __init__(self, x, g=Curve.G):
        self.privkey = x
        self.pubkey = g, g ** x


    @classmethod
    def encrypt(cls, m, pub):
        g, h = pub
        y = randfield(field(g.N)).value
        m_ = map_to_curve(m)
        s = h ** y
        c1 = g ** y
        c2 = s * m_
        return (c1, c2)


    @classmethod
    def decrypt(cls, x, c):
        c1, c2 = c
        s = c1 ** x
        m = (c2 - s)
        return map_from_curve(m)


    def E(self, m):
        return self.encrypt(m, self.pubkey)


    def D(self, c):
        return self.decrypt(self.privkey, c)



"""
DOC: https://eips.ethereum.org/EIPS/eip-2494
"""
import klefki.const as const
from klefki.types.algebra.fields import FiniteField
from klefki.types.algebra.groups import EllipticCurveGroup
from klefki.types.algebra.groups import EllipicCyclicSubgroup
from klefki.curves.arith import short_weierstrass_form_curve_addition2



class FiniteFieldBabyJubjub(FiniteField):
    P = const.BABYJUBJUB_P


class EllipticCurveBabyJubjub(EllipticCurveGroup):
    """
    Twisted Edwards Form (standard)
    y^2 = x^3 + Ax^2 + x
    Montgomery Form
    By^2 = x^3 + A x^2 + x
    """
    A = const.BABYJUBJUB_A
    B = const.BABYJUBJUB_B
    N = const.BABYJUBJUB_N


    def op(self, g):
        # x3 = (x1*y2 + y1*x2)/(1 + b*x1*x2*y1*y2)
        # y3 = (y1*y2 - a*x1*x2)/(1 - b*x1*x2*y1*y2)

        if g.value == 0:
            return self
        if self.value == 0:
            return g
        field = self.value[0].__class__

        m = field(self.B) * self.x * g.x * self.y * g.y

        x3 = (self.x * g.y + self.y * g.x) / (field(1) + m)
        y3 = (self.y * g.y - field(self.A) * self.x * g.x) / (field(1) - m)
        return self.__class__((x3, y3))

EllipticCurveBabyJubjub.G = EllipticCurveBabyJubjub((
        FiniteFieldBabyJubjub(const.BABYJUBJUB_Gx),
        FiniteFieldBabyJubjub(const.BABYJUBJUB_Gy)
))



from klefki.algorithms import bits_little_endian_from_bytes, bytes_from_bits_little_endian


class Sponge:

    def __init__(self, b, r, f=lambda x: x[::-1]):
        assert b % 8 == 0
        assert b - r >= 8
        self.f = f
        self.b = b
        self.r = r
        self.R = [int(i) for i in "0"*r]
        self.C = [int(i) for i in "0"*(b-r)]

    @property
    def S(self):
        return self.R + self.C

    @S.setter
    def S(self, v):
        self.R = v[:self.r]
        self.C = v[self.r:]

    @classmethod
    def padding(cls, dataset: bytes, r: int):
        dataset = [int(i) for i in bits_little_endian_from_bytes(dataset)]
        left = len(dataset) % (r)
        if left != 0:
            dataset = list(dataset) + [0] * (r - left)
        assert len(dataset) % (r) == 0
        return dataset

    def absorbs(self, dataset: bytes):
        dataset = self.padding(dataset, self.r)
        for i in range(0, len(dataset) // self.r):
            m = dataset[i*self.r: i*self.r+self.r]
            for i in range(0, self.r):
                self.R[i] ^= m[i]
            self.S = self.f(self.S)

    def squeezed(self, length):
        ret = []
        for i in range((length // self.r)+1):
            ret = ret + self.R
            self.S = self.f(self.S)
        ret = "".join([str(i) for i in ret])
        return bytes_from_bits_little_endian(ret)[:length]



from typing import Callable
from pyasn1.codec.der.encoder import encode as edr_encode
from pyasn1.codec.der.decoder import decode as edr_decode
import klefki.crypto.ecdsa.secp256k1 as ecdsa
from klefki.asn import signature as sig_asn
from klefki.bitcoin.private import decode_privkey
from klefki.bitcoin.public import decode_pubkey
from klefki.utils import int_to_byte, byte_to_int, b64encode, b58encode
from klefki.types.algebra.concrete import FiniteFieldCyclicSecp256k1 as CF
from klefki.types.algebra.isomorphism import bijection, do


@bijection(edr_encode)
def decode_edr(value):
    return edr_decode(value, asn1Spec=sig_asn.ECDSA_Sig_Value())[0]


sig_encode = do(edr_encode, b58encode)
sig_decode = ~sig_encode


def to_sigtype(sig):
    return tuple([v.value for v in sig])


@bijection(to_sigtype)
def from_sigtype(sig):
    return tuple([CF(v) for v in sig])


def msgsig_to_bytes(v, r, s):
    return chr(v), int_to_byte(r), int_to_byte(s)


@bijection(msgsig_to_bytes)
def msgid_from_bytes(v, r, s):
    return ord(v), byte_to_int(r), byte_to_int(s)


def compose(v, r, s) -> bytes:
    return v + '\x00' * (32 - len(r)) + r + '\X00' * (32 - len(s)) + s


@bijection(compose)
def decompose(b: bytes) -> ecdsa.SigType:
    return b[0], b[1: 33], b[33:]


msgsig_encode: Callable[[ecdsa.SigType], bytes] = do(
    from_sigtype,
    msgsig_to_bytes,
    compose,
    b64encode
)
msgsig_decode: Callable[[bytes], ecdsa.SigType] = ~(msgsig_encode)


def sign(privkey: str, msg: str):
    struct = sig_asn.ECDSA_Sig_Value()
    key = decode_privkey(privkey)
    sig = ecdsa.sign(key, msg)
    struct['r'], struct['s'] = sig[1].value, sig[2].value
    return sig_encode(struct)


def sign_bytes(privkey: str, msg: bytes):
    return sign(privkey, msg.decode())


def verify(pubkey: str, sig: str, msg: str):
    pubkey = decode_pubkey(pubkey)
    sig = sig_decode(sig)
    sig = CF(sig[0]._value), CF(sig[1]._value)
    return ecdsa.verify(pubkey, sig, msg)



"""
ref: https://www.wikiwand.com/zh-hans/%E8%BC%BE%E8%BD%89%E7%9B%B8%E9%99%A4%E6%B3%95
"""
from base64 import b64decode
from klefki.algorithms import extended_euclidean_algorithm
from klefki.numbers import lcm
from klefki.utils import parse_lv_format
from klefki.crypto.pkcs import gen_pad

__all__ = (
    "RSA",
    "mod_inverse",
)


def mod_inverse(a, m):
    """
    gcd = as + mt
    ab = 1 (mod m)
    a and m are relatively prime
    """
    g, s, t = extended_euclidean_algorithm(a, m)
    if g != 1:
        raise Exception('modular inverse does not exist')
    else:
        return s % m


class RSA:
    """
    ref: https://www.wikiwand.com/en/RSA_(cryptosystem)
    You should always use encrypt and decrypt never use
    encrypt/decrypt with xxx-key without knowing what you are doing.
    """
    e = 65537

    def __init__(self, p, q):
        assert p > 0
        assert q > 0
        assert p != q
        self.n = p * q
        lambda_n = lcm((p - 1), (q - 1))
        assert lambda_n > self.e
        self.d = mod_inverse(self.e, lambda_n)

    @property
    def public_key(self):
        return self.n, self.e

    @property
    def private_key(self):
        return self.n, self.d


    def decrypt_with_pub_key(self, block: int):
        return pow(block, self.e, self.n)

    def encrypt_with_private_key(self, block: int) -> int:
        return pow(block, self.d, self.n)

    def encrypt_with_pub_key(self, block: int):
        return pow(block, self.e, self.n)

    def decrypt_with_private_key(self, block: int) -> int:
        return pow(block, self.d, self.n)



    encrypt_block = encrypt_with_pub_key
    decrypt_block = decrypt_with_private_key

    @staticmethod
    def decrypt(msg, e, n):
        return pow(msg, e, n)

    def encrypt_string(self, message: str) -> list:
        return [self.encrypt_with_pub_key(ord(x)) for x in message]

    def decrypt_string(self, encrypted: list):
        return ''.join([chr(self.decrypt_with_private_key(x)) for x in encrypted])

    @staticmethod
    def parse_lv_pubkey(pub):
        pub = b64decode(pub)
        prototol, e, n = parse_lv_format(pub)
        e = int.from_bytes(e, "big")
        n = int.from_bytes(n, "big")
        return e, n

    @classmethod
    def verify_ssh_rsa_sig(cls, sig, msg, pub, alog="SHA256"):
        e, n = cls.parse_lv_pubkey(pub)
        padded_msg = gen_pad(n, msg, alog)
        return cls.decrypt(int(sig, 16), e, n) == int(padded_msg, 16)



'''
* ref https://en.wikipedia.org/wiki/Shamir%27s_Secret_Sharing
The essential idea of Adi Shamir's threshold scheme is that 2 points are sufficient to define a line, 3 points are sufficient to define a parabola, 4 points to define a cubic curve and so forth. That is, it takes $k$ points to define a polynomial of degree $k-1$.



'''

import random
from functools import reduce
from operator import add, mul
from klefki.types.algebra.fields import FiniteField
from klefki.types.algebra.meta import field
from klefki.types.algebra.utils import randfield


class SSSS:
    def __init__(self, F: FiniteField, secret=None, k=None, n=None, poly_params=None):
        self.F = F
        if all([secret, k, n]):
            self.setup(secret, k, n, poly_params)

    def setup(self, secret, k, n, poly_params = []):
        '''
        k: threshold
        '''
        self.k = k
        self.n = n
        if not poly_params:
            poly_params = [randfield(self.F) for _ in range(k-1)]

        self.f = lambda x: self.F(secret) + reduce(
            add, [poly_params[i] * (x ** (i + 1)) for i in range(k-1)])

        return self

    def join(self, x=None):
        assert x != 0
        if not x:
            x = randfield(self.F)
        if not hasattr(self, 'f'):
            raise Exception("Needs to encrypt first")
        return (self.F(x), self.f(x))


    @staticmethod
    def decrypt(shares):
        x, fx = zip(*shares)
        k = len(fx)
        return reduce(add,
                      [fx[j] * reduce(
                          mul,
                          [x[m] / (x[m]-x[j]) for m in range(k-1) if m != j]) for j in range(k-1)])

    @classmethod
    def combine(cls, shares):
        return cls.decrypt(shares)



import klefki.const as const
from klefki.types.algebra.fields import FiniteField
from klefki.types.algebra.groups import EllipticCurveGroup
from klefki.types.algebra.groups import EllipicCyclicSubgroup
from klefki.curves.arith import short_weierstrass_form_curve_addition2


class FiniteFieldSecp256k1(FiniteField):
    P = const.SECP256K1_P


class FiniteFieldCyclicSecp256k1(FiniteField):
    P = const.SECP256K1_N


class EllipticCurveGroupSecp256k1(EllipticCurveGroup):
    """
    y^2 = x^3 + A * x + B
    """

    N = const.SECP256K1_N
    A = const.SECP256K1_A
    B = const.SECP256K1_B

    @property
    def x(self):
        if self == self.identity:
            return 0
        return self.value[0]

    @property
    def y(self):
        if self == self.identity:
            return 0
        return self.value[1]

    def op(self, g):
        field = self.value[0].__class__
        x, y = short_weierstrass_form_curve_addition2(
            self.x, self.y,
            g.x, g.y,
            field(0),
            field(0),
            field(0),
            field(self.A),
            field(self.B),
            field
        )
        if x == y == field(0):
            return self.__class__(0)
        return self.__class__((x, y))

EllipticCurveGroupSecp256k1.G = EllipticCurveGroupSecp256k1(
    (
        FiniteFieldSecp256k1(const.SECP256K1_Gx),
        FiniteFieldSecp256k1(const.SECP256K1_Gy)
    )
)



import hashlib
from klefki.utils import parse_lv_format
# https://medium.com/@bn121rajesh/rsa-sign-and-verify-using-openssl-behind-the-scene-bf3cac0aade2

"""
For the six hash functions mentioned in Appendix B.1, the DER
encoding T of the DigestInfo value is equal to the following:

MD2:     (0x)30 20 30 0c 06 08 2a 86 48 86 f7 0d 02 02 05 00 04
               10 || H.
MD5:     (0x)30 20 30 0c 06 08 2a 86 48 86 f7 0d 02 05 05 00 04
               10 || H.
SHA-1:   (0x)30 21 30 09 06 05 2b 0e 03 02 1a 05 00 04 14 || H.
SHA-256: (0x)30 31 30 0d 06 09 60 86 48 01 65 03 04 02 01 05 00
               04 20 || H.
SHA-384: (0x)30 41 30 0d 06 09 60 86 48 01 65 03 04 02 02 05 00
               04 30 || H.
SHA-512: (0x)30 51 30 0d 06 09 60 86 48 01 65 03 04 02 03 05 00
                  04 40 || H.
"""

SHA1 = "3021300906052b0e03021a05000414"
SHA256 = "3031300d060960864801650304020105000420"


def gen_pad(n, msg, hash_type):
    key_len = len(hex(n)) - 2
    if hash_type == "SHA1":
        m = SHA1
        h = hashlib.sha1(msg.encode()).hexdigest()
    elif hash_type == "SHA256":
        m = SHA256
        h = hashlib.sha256(msg.encode()).hexdigest()
    else:
        raise Exception("Im too lazy to add all the hash type support, please open a pull requrest if you like")
    pad = "f" * (key_len - len(h) - len(m) - 4 - 2)
    return "0001" + pad + "00" + m + h



import abc

class NIZK(abc.ABCMeta):
    @abc.abstractmethod
    def callange(self, *args, **kwargs):
        raise NotImplemented()

    @abc.abstractmethod
    def verify(self, *args, **kwargs):
        raise NotImplemented()


class Sigma(NIZK):

    @abc.abstractmethod
    def proof(self, *args, **kwargs):
        raise NotImplemented()

    @abc.abstractmethod
    def callange(self, *args, **kwargs):
        raise NotImplemented()

    @abc.abstractmethod
    def verify(self, *args, **kwargs):
        raise NotImplemented()



from typing import Iterable, Iterator
from itertools import starmap
from hashlib import sha256
from klefki.utils import concat, trunks
from collections import namedtuple

__all__ = ['MerkleTree']

class MerkleLeaf:
    def __init__(self, value, children=None, parent=None):
        self.value = value
        self.children = children or []
        self.parent = parent

    def __repr__(self):
        return "<MerlkeTree::Leaf(%s)>" % self.value


class MerkleTree:
    def __init__(
        self,
        value,
        is_leaf=False,
        parent=None,
        hash_fn = lambda x, y: concat(sha256(x).digest(), sha256(y).digest())
    ):
        self.hash_fn = hash_fn
        self._value = list(value)
        if not len(self.value) % 2 == 0:
            self.value.append(self.value[-1])
        self.leaves = map(MerkleLeaf, self.value)
        self._build_tree(self.leaves)

    def _build_tree(self, leaves):
        height = 0
        while 1:
            leaves_nl = []
            for (l, r) in trunks(leaves, 2):
                leaf = MerkleLeaf(
                        self.hash_fn(l.value, r.value),
                        children=[l, r]
                    )
                leaves_nl.append(leaf)
                l.parent = leaf
                r.parent = leaf
            leaves = leaves_nl
            height += 1
            if len(leaves) == 1:
                self._root = leaves[0]
                self._height = height
                break

    @property
    def height(self):
        return self._height

    @property
    def root(self):
        return self._root

    @property
    def value(self):
        return self._value



"""
from https://github.com/ethereum/research/blob/master/zksnark/code_to_r1cs.py
ref: https://medium.com/@VitalikButerin/quadratic-arithmetic-programs-from-zero-to-hero-f6d558cea649
"""

import ast
import inspect

if 'arg' not in dir(ast):
    ast.arg = type(None)

def parse(code):
    return ast.parse(code).body

# Takes code of the form
# def foo(arg1, arg2 ...):
#     x = arg1 + arg2
#     y = ...
#     return x + y
# And extracts the inputs and the body, where
# it expects the body to be a sequence of
# variable assignments (variables are immutable;
# can only be set once) and a return statement at the end
def extract_inputs_and_body(code):
    o = []
    if len(code) != 1 or not isinstance(code[0], ast.FunctionDef):
        raise Exception("Expecting function declaration")
    # Gather the list of input variables
    inputs = []
    for arg in code[0].args.args:
        if isinstance(arg, ast.arg):
            assert isinstance(arg.arg, str)
            inputs.append(arg.arg)
        elif isinstance(arg, ast.Name):
            inputs.append(arg.id)
        else:
            raise Exception("Invalid arg: %r" % ast.dump(arg))
    # Gather the body
    body = []
    returned = False
    for c in code[0].body:
        if not isinstance(c, (ast.Assign, ast.Return)):
            raise Exception("Expected variable assignment or return")
        if returned:
            raise Exception("Cannot do stuff after a return statement")
        if isinstance(c, ast.Return):
            returned = True
        body.append(c)
    return inputs, body

# Convert a body with potentially complex expressions into
# simple expressions of the form x = y or x = y * z
def flatten_body(body):
    o = []
    for c in body:
        o.extend(flatten_stmt(c))
    return o

# Generate a dummy variable
next_symbol = [0]
def mksymbol():
    next_symbol[0] += 1
    return 'sym_'+str(next_symbol[0])

# "Flatten" a single statement into a list of simple statements.
# First extract the target variable, then flatten the expression
def flatten_stmt(stmt):
    # Get target variable
    if isinstance(stmt, ast.Assign):
        assert len(stmt.targets) == 1 and isinstance(stmt.targets[0], ast.Name)
        target = stmt.targets[0].id
    elif isinstance(stmt, ast.Return):
        target = '~out'
    # Get inner content
    return flatten_expr(target, stmt.value)

# Main method for flattening an expression
def flatten_expr(target, expr):
    # x = y
    if isinstance(expr, ast.Name):
        return [['set', target, expr.id]]
    # x = 5
    elif isinstance(expr, ast.Num):
        return [['set', target, expr.n]]
    # x = y (op) z
    # Or, for that matter, x = y (op) 5
    elif isinstance(expr, ast.BinOp):
        if isinstance(expr.op, ast.Add):
            op = '+'
        elif isinstance(expr.op, ast.Mult):
            op = '*'
        elif isinstance(expr.op, ast.Sub):
            op = '-'
        elif isinstance(expr.op, ast.Div):
            op = '/'
        # Exponentiation gets compiled to repeat multiplication,
        # requires constant exponent
        elif isinstance(expr.op, ast.Pow):
            assert isinstance(expr.right, ast.Num)
            if expr.right.n == 0:
                return [['set', target, 1]]
            elif expr.right.n == 1:
                return flatten_expr(target, expr.left)
            else: # This could be made more efficient via square-and-multiply but oh well
                if isinstance(expr.left, (ast.Name, ast.Num)):
                    nxt = base = expr.left.id if isinstance(expr.left, ast.Name) else expr.left.n
                    o = []
                else:
                    nxt = base = mksymbol()
                    o = flatten_expr(base, expr.left)
                for i in range(1, expr.right.n):
                    latest = nxt
                    nxt = target if i == expr.right.n - 1 else mksymbol()
                    o.append(['*', nxt, latest, base])
                return o
        else:
            raise Exception("Bad operation: " % ast.dump(stmt.op))
        # If the subexpression is a variable or a number, then include it directly
        if isinstance(expr.left, (ast.Name, ast.Num)):
            var1 = expr.left.id if isinstance(expr.left, ast.Name) else expr.left.n
            sub1 = []
        # If one of the subexpressions is itself a compound expression, recursively
        # apply this method to it using an intermediate variable
        else:
            var1 = mksymbol()
            sub1 = flatten_expr(var1, expr.left)
        # Same for right subexpression as for left subexpression
        if isinstance(expr.right, (ast.Name, ast.Num)):
            var2 = expr.right.id if isinstance(expr.right, ast.Name) else expr.right.n
            sub2 = []
        else:
            var2 = mksymbol()
            sub2 = flatten_expr(var2, expr.right)
        # Last expression represents the assignment; sub1 and sub2 represent the
        # processing for the subexpression if any
        return sub1 + sub2 + [[op, target, var1, var2]]
    else:
        raise Exception("Unexpected statement value: %r" % stmt.value)

# Adds a variable or number into one of the vectors; if it's a variable
# then the slot associated with that variable is set to 1, and if it's
# a number then the slot associated with 1 gets set to that number
def insert_var(arr, varz, var, used, reverse=False):
    if isinstance(var, str):
        if var not in used:
            raise Exception("Using a variable before it is set!")
        arr[varz.index(var)] += (-1 if reverse else 1)
    elif isinstance(var, int):
        arr[0] += var * (-1 if reverse else 1)

# Maps input, output and intermediate variables to indices
def get_var_placement(inputs, flatcode):
    return ['~one'] + [x for x in inputs] + ['~out'] + [c[1] for c in flatcode if c[1] not in inputs and c[1] != '~out']


# Convert the flattened code generated above into a rank-1 constraint system
def flatcode_to_r1cs(inputs, flatcode):
    varz = get_var_placement(inputs, flatcode)
    A, B, C = [], [], []
    used = {i: True for i in inputs}
    for x in flatcode:
        a, b, c = [0] * len(varz), [0] * len(varz), [0] * len(varz)
        if x[1] in used:
            raise Exception("Variable already used: %r" % x[1])
        used[x[1]] = True
        if x[0] == 'set':
            a[varz.index(x[1])] += 1
            insert_var(a, varz, x[2], used, reverse=True)
            b[0] = 1
        elif x[0] == '+' or x[0] == '-':
            c[varz.index(x[1])] = 1
            insert_var(a, varz, x[2], used)
            insert_var(a, varz, x[3], used, reverse=(x[0] == '-'))
            b[0] = 1
        elif x[0] == '*':
            c[varz.index(x[1])] = 1
            insert_var(a, varz, x[2], used)
            insert_var(b, varz, x[3], used)
        elif x[0] == '/':
            insert_var(c, varz, x[2], used)
            a[varz.index(x[1])] = 1
            insert_var(b, varz, x[3], used)
        A.append(a)
        B.append(b)
        C.append(c)
    return A, B, C

# Get a variable or number given an existing input vector
def grab_var(varz, assignment, var):
    if isinstance(var, str):
        return assignment[varz.index(var)]
    elif isinstance(var, int):
        return var
    else:
        raise Exception("What kind of expression is this? %r" % var)

# Goes through flattened code and completes the input vector
def assign_variables(inputs, input_vars, flatcode):
    varz = get_var_placement(inputs, flatcode)
    assignment = [0] * len(varz)
    assignment[0] = 1
    for i, inp in enumerate(input_vars):
        assignment[i + 1] = inp
    for x in flatcode:
        if x[0] == 'set':
            assignment[varz.index(x[1])] = grab_var(varz, assignment, x[2])
        elif x[0] == '+':
            assignment[varz.index(x[1])] = grab_var(varz, assignment, x[2]) + grab_var(varz, assignment, x[3])
        elif x[0] == '-':
            assignment[varz.index(x[1])] = grab_var(varz, assignment, x[2]) - grab_var(varz, assignment, x[3])
        elif x[0] == '*':
            assignment[varz.index(x[1])] = grab_var(varz, assignment, x[2]) * grab_var(varz, assignment, x[3])
        elif x[0] == '/':
            assignment[varz.index(x[1])] = grab_var(varz, assignment, x[2]) / grab_var(varz, assignment, x[3])
    return assignment


def code_to_r1cs_with_inputs(code, input_vars):
    inputs, body = extract_inputs_and_body(parse(code))
    flatcode = flatten_body(body)
    A, B, C = flatcode_to_r1cs(inputs, flatcode)
    r = assign_variables(inputs, input_vars, flatcode)
    return r, A, B, C

def mul(a, b):
    return sum(list(map(lambda x: x[0] * x[1], zip(a, b))))

class R1CS:
    @staticmethod
    def parse(code, input_vals):
        s, A, B, C = code_to_r1cs_with_inputs(code, input_vals)
        return (s, A, B, C)

    @staticmethod
    def verify(s, A, B, C):
        ret = True
        for i in range(len(s)-2):
            ret = ret and mul(A[i], s) * mul(B[i], s) == mul(C[i], s)
        return ret

    def r1cs(f):
        src = inspect.getsource(f)
        def code_to_r1cs(*args):
            return code_to_r1cs_with_inputs(src, list(args))
        f.r1cs = code_to_r1cs
        f.src = src
        return f



import sys
from klefki.blockchain import bitcoin, eos, ethereum


# klefki.bitcoin/eos/ethereum was mv to klefki.blockchain
sys.modules['klefki.bitcoin'] = bitcoin
sys.modules['klefki.eos'] = eos
sys.modules['klefki.ethereum'] = ethereum



from functools import wraps, partial
import sys


def command(fn, name=None):

    if not hasattr(command, 'registered'):
        command.registered = {}
    fname = name or fn.__name__
    command.registered[fname] = fn
    return fn


@command
def shell():
    '''
    default Shell
    '''
    import random
    from klefki.utils import to_sha256int
    from klefki.types.algebra.concrete import (
        JacobianGroupSecp256k1 as JG,
        EllipticCurveCyclicSubgroupSecp256k1 as CG,
        EllipticCurveGroupSecp256k1 as ECG,
        FiniteFieldCyclicSecp256k1 as CF
    )
    import IPython

    N = CG.N
    G = CG.G
    IPython.embed()


@command
def help():
    '''
    This help command
    '''
    for k, v in command.registered.items():
        print(k, v.__doc__)


def not_found(cmd):
    print('cmd %s not found' % cmd)


def output(fn):
    @wraps(fn)
    def _(*args, **kwargs):
        return print(fn(*args, **kwargs))
    return _


def router(cmds):
    'fn a b c=1 d=2'
    if isinstance(cmds, str):
        cmds = cmds.split(' ')
    try:
        fn = command.registered.get(cmds[0], None)
    except IndexError as _:
        return help()
    if not fn:
        return not_found(cmds)
    args = [c for c in cmds[1:] if '=' not in c]
    kwargs = dict([c.split('=') for c in cmds[1:] if '=' in c])
    return fn(*args, **kwargs)


main = partial(router, cmds=sys.argv[1:])


if __name__ == "__main__":
    main()



"""
Ref: https://byt3bit.github.io/primesym/mimc/
Ref: https://eprint.iacr.org/2016/492.pdf
"""

from klefki.curves.baby_jubjub import FiniteFieldBabyJubjub as F
from klefki.types.algebra.utils import randfield
from klefki.utils import to_sha256int
from functools import partial, reduce
from itertools import count
import time
from hashlib import sha256

class MiMC:
    @staticmethod
    def F(x, k, c):
        return (x + k + c) ** 3

    def __init__(self, field, r):
        self.field = field
        self.c = [field(0)] + [randfield(field) for i in range(1, r)]
        self.r = r

    def encrypt(self, x, k, r=None):
        if not r: r = self.r
        Fs = [partial(self.F, k=k, c=c) for c in self.c[:r]]
        return reduce(lambda x, y: x + y(x), Fs[1:], Fs[0](x)) + k

    def E(self, *args, **kwargs):
        return self.encrypt(*args, **kwargs)

    @property
    def constants(self):
        return self.c


class FeistelMiMC(MiMC):

    @staticmethod
    def F(x, y, k, c):
        return (y, x + (y + k + c) ** 3)

    def encrypt(self, x, y, k, r=None):
        if not r: r = self.r
        Ks = [(i + self.field(1)) * k for i in range(0, r)]
        Fs = [partial(self.F, k=k, c=c) for (k, c) in zip(Ks[:r], self.c[:r])]
        return reduce(
            lambda x, y: y(*x), Fs[1:], Fs[0](x, y)
        ) + (k, k)
